import { TokenizeMethod, Eat } from '../tokenizer'
import { RemarkParser } from '../RemarkParser'
import { Node, Point, Parent } from 'typed-unist'
import { isWhitespaceCharacter } from 'typed-string-utils'

const C_BACKSLASH = '\\'
const C_TICK = '`'
const C_DASH = '-'
const C_PIPE = '|'
const C_COLON = ':'
const C_SPACE = ' '
const C_NEWLINE = '\n'
const C_TAB = '\t'

const MIN_TABLE_COLUMNS = 1
const MIN_TABLE_ROWS = 2

const TABLE_ALIGN_LEFT = 'left'
const TABLE_ALIGN_CENTER = 'center'
const TABLE_ALIGN_RIGHT = 'right'
const TABLE_ALIGN_NONE: null = null

export const table: TokenizeMethod = function (this: RemarkParser, eat: Eat, value: string, silent?: boolean): Node | boolean {
  const self = this
  let index
  let alignments: string
  let alignment
  let subvalue
  let row
  let length
  let lines
  let queue
  let character
  let hasDash
  let align
  let cell
  let preamble
  let count
  let opening
  let now: Point
  let position
  let lineCount
  let line
  let rows: Parent[]
  let tableNode: Parent
  let lineIndex
  let pipeIndex
  let first

  /* Exit when not in gfm-mode. */
  if (!self.options.gfm) {
    return
  }

  /* Get the rows.
   * Detecting tables soon is hard, so there are some
   * checks for performance here, such as the minimum
   * number of rows, and allowed characters in the
   * alignment row. */
  index = 0
  lineCount = 0
  length = value.length + 1
  lines = []

  while (index < length) {
    lineIndex = value.indexOf(C_NEWLINE, index)
    pipeIndex = value.indexOf(C_PIPE, index + 1)

    if (lineIndex === -1) {
      lineIndex = value.length
    }

    if (pipeIndex === -1 || pipeIndex > lineIndex) {
      if (lineCount < MIN_TABLE_ROWS) {
        return
      }

      break
    }

    lines.push(value.slice(index, lineIndex))
    lineCount++
    index = lineIndex + 1
  }

  /* Parse the alignment row. */
  subvalue = lines.join(C_NEWLINE)
  alignments = lines.splice(1, 1)[0] || ''
  index = 0
  length = alignments.length
  lineCount--
  alignment = false
  align = []

  while (index < length) {
    character = alignments.charAt(index)

    if (character === C_PIPE) {
      hasDash = null

      if (alignment === false) {
        if (first === false) {
          return
        }
      } else {
        align.push(alignment)
        alignment = false
      }

      first = false
    } else if (character === C_DASH) {
      hasDash = true
      alignment = alignment || TABLE_ALIGN_NONE
    } else if (character === C_COLON) {
      if (alignment === TABLE_ALIGN_LEFT) {
        alignment = TABLE_ALIGN_CENTER
      } else if (hasDash && alignment === TABLE_ALIGN_NONE) {
        alignment = TABLE_ALIGN_RIGHT
      } else {
        alignment = TABLE_ALIGN_LEFT
      }
    } else if (!isWhitespaceCharacter(character)) {
      return
    }

    index++
  }

  if (alignment !== false) {
    align.push(alignment)
  }

  /* Exit when without enough columns. */
  if (align.length < MIN_TABLE_COLUMNS) {
    return
  }

  /* istanbul ignore if - never used (yet) */
  if (silent) {
    return true
  }

  /* Parse the rows. */
  position = -1
  rows = []

  tableNode = eat(subvalue).reset({
    type: 'table',
    align,
    children: rows,
  } as Node) as Parent

  while (++position < lineCount) {
    line = lines[position]
    row = {type: 'tableRow', children: []} as Node

    /* Eat a newline character when this is not the
     * first row. */
    if (position) {
      eat(C_NEWLINE)
    }

    /* Eat the row. */
    eat(line).reset(row, tableNode)

    length = line.length + 1
    index = 0
    queue = ''
    cell = ''
    preamble = true
    count = null
    opening = null

    while (index < length) {
      character = line.charAt(index)

      if (character === C_TAB || character === C_SPACE) {
        if (cell) {
          queue += character
        } else {
          eat(character)
        }

        index++
        continue
      }

      if (character === '' || character === C_PIPE) {
        if (preamble) {
          eat(character)
        } else {
          if (character && opening) {
            queue += character
            index++
            continue
          }

          if ((cell || character) && !preamble) {
            subvalue = cell

            if (queue.length > 1) {
              if (character) {
                subvalue += queue.slice(0, queue.length - 1)
                queue = queue.charAt(queue.length - 1)
              } else {
                subvalue += queue
                queue = ''
              }
            }

            now = eat.now()

            eat(subvalue)({
              type: 'tableCell',
              children: self.tokenizeInline(cell, now),
            } as Node, row as Parent)
          }

          eat(queue + character)

          queue = ''
          cell = ''
        }
      } else {
        if (queue) {
          cell += queue
          queue = ''
        }

        cell += character

        if (character === C_BACKSLASH && index !== length - 2) {
          cell += line.charAt(index + 1)
          index++
        }

        if (character === C_TICK) {
          count = 1

          while (line.charAt(index + 1) === character) {
            cell += character
            index++
            count++
          }

          if (!opening) {
            opening = count
          } else if (count >= opening) {
            opening = 0
          }
        }
      }

      preamble = false
      index++
    }

    /* Eat the alignment row. */
    if (!position) {
      eat(C_NEWLINE + alignments)
    }
  }

  return tableNode
} as TokenizeMethod
